#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Commands for Pyppeteer."""
import argparse

from biolib.pyppeteer.pyppeteer.browser_fetcher import BrowserFetcher


def install() -> None:
    """Download and Chromium/Firefox to specified folder"""
    #parser = argparse.ArgumentParser(description=install.__doc__)
    #parser.add_argument('-r', '--revision', action="store", type=str, default=None)
    #parser.add_argument('-p', '--product', action="store", type=str, default=None)
    #parser.add_argument('-l', '--location', action="store")
    #parsed = parser.parse_args()
    dl = BrowserFetcher(None) #parsed.location)
    dl.download(None) #parsed.revision)


if __name__ == '__main__':
    install()



import asyncio
import json
import logging
import sys
from typing import Any, Awaitable, Dict

import websockets
from pyee import AsyncIOEventEmitter

from biolib.pyppeteer.pyppeteer.errors import NetworkError
from biolib.pyppeteer.pyppeteer.events import Events
from biolib.pyppeteer.pyppeteer.websocket_transport import WebsocketTransport

if sys.version_info < (3, 8):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger_connection = logging.getLogger(__name__ + '.Connection')


class TargetInfo(TypedDict, total=False):
    type: str


class MessageParams(TypedDict, total=False):
    targetInfo: TargetInfo
    sessionId: str


class MessageError(TypedDict, total=False):
    message: str
    data: Any


class Message(TypedDict, total=False):
    method: str
    id: int
    params: MessageParams
    error: MessageError
    result: Any


class Connection(AsyncIOEventEmitter):
    """Connection management class."""

    def __init__(
        self, url: str, transport: WebsocketTransport, delay: float = 0, loop: asyncio.AbstractEventLoop = None,
    ) -> None:
        """Make connection.

        :arg str url: WebSocket url to connect devtool.
        :arg int delay: delay to wait before processing received messages.
        """
        super().__init__()
        self._url = url
        self._lastId = 0
        self._callbacks: Dict[int, asyncio.Future] = {}
        self._delay = delay / 1000

        self._transport = transport

        self.loop = loop or asyncio.get_event_loop()
        self._sessions: Dict[str, CDPSession] = {}
        self._connected = False
        self._closed = False
        self.loop.create_task(self._recv_loop())

    @staticmethod
    def fromSession(session: 'CDPSession') -> 'Connection':
        return session._connection

    def session(self, sessionId) -> 'CDPSession':
        return self._sessions.get(sessionId)

    @property
    def url(self) -> str:
        """Get connected WebSocket url."""
        return self._url

    async def _recv_loop(self) -> None:
        try:
            self._connected = True
            self.connection = self._transport
            self.connection.onmessage = lambda msg: self._onMessage(msg)
            self.connection.onclose = self._onClose
            while self._connected:
                try:
                    await self.connection.recv()
                except (websockets.ConnectionClosed, ConnectionResetError) as excpt:
                    logger.warning(f'Transport connection closed: {excpt}')
                    break
                # wait 1 async loop frame, no other data will be accessible in between frames
                await asyncio.sleep(0)
        except Exception as excpt:
            await self.dispose(reason=str(excpt))
            raise excpt
        await self.dispose(reason=None)

    async def _async_send(self, msg: Message) -> None:
        while not self._connected:
            await asyncio.sleep(self._delay)
        try:
            remove_none_items_inplace(msg)
            msg_to_send = json.dumps(msg)
            await self.connection.send(msg_to_send)
            logger_connection.debug(f'SEND ▶ {msg_to_send}')
        except websockets.ConnectionClosed:
            logger.error('connection unexpectedly closed')
            callback = self._callbacks.get(msg['id'], None)
            if callback and not callback.done():
                callback.set_result(None)
                await self.dispose()

    def send(self, method: str, params: dict = None) -> Awaitable:
        """Send message via the connection."""
        # Detect connection availability from the second transmission
        if self._lastId and not self._connected:
            raise ConnectionError('Connection is closed')
        id_ = self._rawSend({'method': method, 'params': params or {}})
        callback = self.loop.create_future()
        callback.error: Exception = NetworkError()  # type: ignore
        callback.method: str = method  # type: ignore
        self._callbacks[id_] = callback
        return callback

    def _rawSend(self, message: Message) -> int:
        self._lastId += 1
        id_ = self._lastId
        message['id'] = id_
        self.loop.create_task(self._async_send(message))
        return id_

    async def _onMessage(self, msg: str) -> None:
        loaded_msg: Message = json.loads(msg)
        if self._delay:
            await asyncio.sleep(self._delay)
        logger_connection.debug(f'◀ RECV {loaded_msg}')

        # Handle Target attach/detach methods
        if loaded_msg.get('method') == 'Target.attachedToTarget':
            sessionId = loaded_msg['params']['sessionId']
            self._sessions[sessionId] = CDPSession(
                connection=self,
                targetType=loaded_msg['params']['targetInfo']['type'],
                sessionId=sessionId,
                loop=self.loop,
            )
        elif loaded_msg.get('method') == 'Target.detachedFromTarget':
            session = self._sessions.get(loaded_msg['params']['sessionId'])
            if session:
                session._onClosed()
                del self._sessions[loaded_msg['params']['sessionId']]

        if loaded_msg.get('sessionId'):
            session = self._sessions.get(loaded_msg['sessionId'])
            if session:
                session._onMessage(loaded_msg)
        elif loaded_msg.get('id'):
            # Callbacks could be all rejected if someone has called `.dispose()`
            callback = self._callbacks.get(loaded_msg['id'])
            if callback:
                if loaded_msg.get('error'):
                    callback.set_exception(createProtocolError(callback.error, callback.method, loaded_msg))
                else:
                    callback.set_result(loaded_msg.get('result'))
                del self._callbacks[loaded_msg['id']]
        else:
            self.emit(loaded_msg['method'], loaded_msg['params'])

    async def _onClose(self) -> None:
        if self._closed:
            return
        self._closed = True
        self._transport.onmessage = None
        self._transport.onclose = None

        for cb in self._callbacks.values():
            cb.set_exception(
                rewriteError(
                    cb.error,  # type: ignore
                    f'Protocol error {cb.method}: Target closed.',  # type: ignore
                )
            )
        self._callbacks.clear()

        for session in self._sessions.values():
            session._onClosed()
        self._sessions.clear()

        # close connection
        if hasattr(self, 'connection'):  # may not have connection
            await self.connection.close()
        self._sessions.clear()
        self.emit(Events.Connection.Disconnected)

    async def dispose(self, code: int = 1000, reason: str = None) -> None:
        """Close all connection."""
        self._connected = False
        await self._onClose()
        await self._transport.close(code=code, reason=str(reason))

    async def createSession(self, targetInfo: Dict) -> 'CDPSession':
        """Create new session."""
        resp = await self.send('Target.attachToTarget', {'targetId': targetInfo['targetId'], 'flatten': True})
        sessionId = resp.get('sessionId')
        return self._sessions[sessionId]


def createProtocolError(error: Exception, method: str, obj: Dict) -> Exception:
    message = f'Protocol error ({method}): {obj["error"]["message"]}'
    if 'data' in obj['error']:
        message += f' {obj["error"]["data"]}'
    return rewriteError(error, message)


def rewriteError(error: Exception, message: str) -> Exception:
    error.args = (message,)
    return error


def remove_none_items_inplace(o: Dict[str, Any]) -> None:
    """
    Removes items that have a value of None. There are instances in puppeteer where a object (dict) is sent which has
    undefined values, which are then omitted from the resulting json. This function emulates such behaviour, removing
    all k:v pairs where v = None
    :param o:
    :return Dict[str, Any]: dict without any None values
    """
    none_keys = []
    for key, value in o.items():
        if isinstance(value, dict):
            remove_none_items_inplace(value)
        if value is None:
            none_keys.append(key)
    for key in none_keys:
        del o[key]


from biolib.pyppeteer.pyppeteer.connection.cdpsession import CDPSession  # isort:skip



"""
CLI script to generate Protocol types
"""
import asyncio
import json
import logging
import re
import time
from argparse import ArgumentParser
from datetime import datetime
from functools import partial
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, Hashable, List, Literal, Match, Tuple, Union

import networkx as nx

from pyppeteer import launch

handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('[{levelname}] {name}: {message}', style='{'))
logging.getLogger('pyppeteer').addHandler(handler)

logger = logging.getLogger('CLI')
logger.addHandler(handler)
logger.setLevel(logging.INFO)
handler.setLevel(logging.INFO)


class ProtocolTypesGenerator:
    """
    Class for generating a black formatter-semi-compliant code relating relating to the CDP specification.
    Attributes:
        td_references: set of relations of one TypedDict to another. Needed to detect/resolve recursive references
        domains: List containing all the domains of the CDP protocol
        all_known_types: Dict containing all known types. Needed to resolve forward references.
        typed_dicts: Dict of typed_dict_name to TypedDictGenerator(). We need a dict so we can access arbitrary elements
            and update them if a recursive reference is found
        code_gen = instance of TypingCodeGenerator for recording code lines
        _extern_code_gen = instance of TypingCodeGenerator for actually recording code the will appear outside the if
            TYPE_CHECKING guard
    """

    _forward_ref_re = r'\'Protocol\.(\w+\.\w+)\''
    js_to_py_types = {
        'any': 'Any',
        'string': 'str',
        'object': 'Dict[str, str]',
        'boolean': 'bool',
        'number': 'float',
        'integer': 'int',
        'binary': 'bytes',
    }

    def __init__(self):
        self.td_references = set()
        self.domains = []
        self.all_known_types = {}
        self.typed_dicts = {}
        self.code_gen = TypingCodeGenerator()
        self._extern_code_gen = TypingCodeGenerator()

    def _resolve_forward_ref_re_sub_repl(self, match: Match, fw_ref: bool) -> str:
        domain_, ref = match.group(1).split('.')
        resolved_fwref = self.all_known_types[domain_][ref]

        # resolve nested forward references
        if re.search(self._forward_ref_re, resolved_fwref):
            resolved_fwref = self.resolve_forward_ref_on_line(resolved_fwref)
        if (
            fw_ref
            and resolved_fwref not in self.js_to_py_types.values()
            and not resolved_fwref.startswith('Literal')
            and not resolved_fwref.startswith('List')
        ):
            # forward ref to a typed dict, not sure that it will be defined
            resolved_fwref = f'\'{resolved_fwref}\''

        return resolved_fwref

    def resolve_forward_ref_on_line(self, line: str, fw_ref: bool = True) -> str:
        """
        Replaces a forward reference in the form 'Protocol.domain.ref' to the actual value of Protocol.domain.ref
        Args:
            line: line in which protocol forward reference occurs.
            fw_ref: whether or not to forward reference the resolved reference
        Returns:
            str: line with resolved forward reference
        """
        # PyCharm can't handle partial
        # noinspection PyTypeChecker
        return re.sub(self._forward_ref_re, partial(self._resolve_forward_ref_re_sub_repl, fw_ref=fw_ref), line)

    async def _retrieve_top_level_domain(self):
        browser = await launch(args=['--no-sandbox', '--disable-setuid-sandbox'])
        base_endpoint = re.search(r'ws://([0-9A-Za-z:.]*)/', browser.wsEndpoint).group(1)
        page = await browser.newPage()

        logger.info(f'Loading raw protocol specification')
        t_start = time.perf_counter()

        await page.goto(f'http://{base_endpoint}/json/protocol')
        page_content = await page.evaluate('document.documentElement.innerText')
        try:
            await browser.close()
        except Exception as e:
            logger.warning(f'Exception on browser close: {e}')

        logger.info(f'Loading raw protocol specification in {time.perf_counter()-t_start:.2f}s')
        self.domains = json.loads(page_content)['domains']

    def retrieve_top_level_domain(self):
        """
        Fetches and data, parses it, and sets the class variable 'domains' to it for later use.
        Returns: None
        """
        asyncio.get_event_loop().run_until_complete(self._retrieve_top_level_domain())

    def gen_spec(self):
        """
        Generate the Protocol class file lines within self.code_gen attribute. Uses an IndentManager context manager to
        keep track of the current indentation level. Resolves all forward references. Expands recursive types to an
        approximation of the cyclic type referenced. Wraps protocol class in a if TYPE_CHECKING guard and provides a
        dummy protocol class for runtime for performance.

        Returns: None
        """

        self.generate_header_doc_string()

        logger.info(f'Generating protocol spec')
        t_start = time.perf_counter()

        self.code_gen.add_code('class Protocol:')
        with self.code_gen.indent_manager:
            for domain in self.domains:
                domain_name = domain['domain']
                self.code_gen.add_code(f'class {domain_name}:')
                self.all_known_types[domain_name] = {}
                with self.code_gen.indent_manager:
                    self.code_gen.add_comment_from_info(domain)
                    for type_info in domain.get('types', []):
                        self.add_type_item(type_info, 'id', 'properties', domain_name)

                    for payload in domain.get('events', []):
                        payload["name"] = payload["name"] + 'Payload'
                        self.add_type_item(payload, 'name', 'parameters', domain_name, type_conversion_fallback='None')

                    for command_info in domain.get('commands', []):
                        for td_key, suffix in (('properties', 'Parameters'), ('returns', 'ReturnValues')):
                            command_info['name'] = re.subn(r'(Parameters$|$)', suffix, command_info['name'], 1)[0]
                            self.add_type_item(
                                command_info, 'name', td_key, domain_name, type_conversion_fallback='None'
                            )
                self.code_gen.add_newlines(num=2)

            self.generate_overview()
            self.resolve_all_fw_refs()

        self.expand_recursive_references()
        self.typed_dicts = {k: v for k, v in sorted(self.typed_dicts.items(), key=lambda x: x[0])}
        # all typed dicts are inserted prior to the main Protocol class
        for index, td in enumerate(self.typed_dicts.values()):
            td.add_newlines(num=1)
            self.code_gen.add_code(lines=td.code_lines, lines_classification='inserted')

        self.code_gen.add_code(lines=self._extern_code_gen.code_lines)
        logger.info(f'Generated protocol spec in {time.perf_counter() - t_start:.2f}s')

    def generate_header_doc_string(self):
        """
        Generates a headers doc string for the top of the file.
        Returns: None
        """
        self.code_gen.add_code('"""', lines_classification='inserted')
        self.code_gen.add_code(
            f'''\
            Automatically generated by ./{Path(__file__).relative_to(Path(__file__).parents[1]).as_posix()}
            Attention! This file should *not* be modified directly! Instead, use the script to update it.

            Last regeneration: {datetime.utcnow()}''',
            lines_classification='inserted',
        )
        self.code_gen.add_code('"""', lines_classification='inserted')

    def resolve_all_fw_refs(self) -> None:
        """
        Resolves all forward reference to the root of said references eg 'Protocol.Animation.thingParams' -> thingParams
        Returns: None
        """
        # no need for copying list as we aren't adding/removing elements
        # resolve forward refs in main protocol class
        for index, line in enumerate(self.code_gen.code_lines):
            # skip empty lines or lines positively without forward reference
            if not line.strip() or 'Protocol' not in line:
                continue
            self.code_gen.code_lines[index] = self.resolve_forward_ref_on_line(line, fw_ref=False)
        # resolve forward refs in typed dicts, and store instances where s TypedDict references another
        for td_name, td in self.typed_dicts.items():
            for index, line in enumerate(td.code_lines):
                resolved_fw_ref = self.resolve_forward_ref_on_line(line)
                resolved_fw_ref_splits = resolved_fw_ref.split(': ')
                if len(resolved_fw_ref_splits) == 2:  # only pay attention to actual resolve fw refs
                    ref = resolved_fw_ref_splits[1]
                    td_ref_re = r'^(?:List\[)?\'(\w+)\'\]?'
                    if re.search(td_ref_re, ref):
                        self.td_references.add((td_name, re.search(td_ref_re, ref).group(1)))

                self.typed_dicts[td_name].code_lines[index] = resolved_fw_ref

    def generate_overview(self) -> None:
        """
        Generate several convenience overview classes, listed in overview_info
        Returns: None
        """
        overview_info = {
            'Events': ('events', '\'{domain}.{item_name}\'', True, False),
            'CommandParameters': ('commands', '\'Protocol.{domain}.{item_name}\'', False, False),
            'CommandReturnValues': ('commands', '\'Protocol.{domain}.{item_name}\'', False, False),
            'CommandNames': ('commands', '\'{domain}.{item_name}\'', True, True),
        }

        last_overview_class_name = [*overview_info.keys()][-1]
        for overview_class_name, (domain_key, item_fmt, gen_externally, no_suffix) in overview_info.items():
            overview_code_gen = TypingCodeGenerator(init_imports=False)
            overview_code_gen.add_code(f'class {overview_class_name}:')
            with overview_code_gen.indent_manager:
                for domain in self.domains:
                    if domain_key in domain:
                        overview_code_gen.add_code(f'class {domain["domain"]}:')
                        with overview_code_gen.indent_manager:
                            for item in domain[domain_key]:
                                if no_suffix:
                                    item['name'] = re.sub('(ReturnValues|Parameters)$', '', item['name'])
                                formatted_name = item_fmt.format(domain=domain['domain'], item_name=item['name'])
                                overview_code_gen.add_code(f'{item["name"]} = {formatted_name}')
                        overview_code_gen.add_newlines(num=1)
            if overview_class_name != last_overview_class_name:
                # don't add newlines to EOF
                overview_code_gen.add_newlines(num=1)
            if gen_externally:
                self._extern_code_gen.add_code(lines=overview_code_gen.code_lines)
            else:
                self.code_gen.add_code(lines=overview_code_gen.code_lines)

    def add_type_item(
        self,
        type_info,
        type_name_key,
        td_key,
        domain_name,
        type_conversion_fallback: Union[str, Literal[False]] = False,
    ) -> None:
        """
        Adds a class attr based on type_info, type_name_key, td_key, domain_name, and type_conversion_fallback

        Args:
            type_info: Dict containing info pertaining to the type
            type_name_key: Key to the name of the type
            td_key: Key to the item which contains TypedDict info
            domain_name: Name of domain
            type_conversion_fallback: If not false, used when the type cannot be determined from type_info

        Returns: None
        """
        self.code_gen.add_comment_from_info(type_info)
        item_name = type_info[type_name_key]
        if td_key in type_info:
            td = self.generate_typed_dict(type_info, domain_name)
            self.typed_dicts.update(td)
            type_ = [*td.keys()][0]
            typed_dict = True
        else:
            try:
                type_ = self.convert_js_to_py_type(type_info, domain_name)
            except KeyError:
                if type_conversion_fallback is not False:
                    type_ = type_conversion_fallback
                else:
                    raise
            typed_dict = False

        self.all_known_types[domain_name][item_name] = type_
        if typed_dict:
            # https://github.com/python/mypy/issues/7866
            type_ = f'Union[{type_}]'
        self.code_gen.add_code(f'{item_name} = {type_}')

    def expand_recursive_references(self) -> None:
        """
        Expands recursive references to TypedDicts with Dict[str, Union[Dict[str, Any], str, bool, int, float, List]],
        and adds a comment with the actual type reference.
        Returns: None
        """
        expansion = 'Dict[str, Union[Dict[str, Any], str, bool, int, float, List]]'
        # todo: networkx will soon support sets: https://github.com/networkx/networkx/pull/3907
        for recursive_refs in nx.simple_cycles(nx.DiGraph([*self.td_references])):
            any_recursive_ref = "|".join(recursive_refs)
            for recursing_itm in recursive_refs:
                self.typed_dicts[recursing_itm].filter_lines(
                    (rf'(\s+)(\w+): [\w\[]*?\'({any_recursive_ref})\'\]?', rf'\1# actual: \3\n\1\2: {expansion}'),
                )

    def write_generated_code(self, path: Path) -> None:
        """
        Write generated code lines to the specified path. Writes to a temporary file and checks that file with mypy to
        'resolve' any cyclic references.

        Args:
            path: path to write type code to.

        Returns: None
        """
        if path.is_dir():
            path /= '_protocol.py'
        logger.info(f'Writing generated protocol code to {path}')
        # PyCharm can't handle this path type properly
        # noinspection PyTypeChecker
        with open(path, 'w') as p:
            p.write(str(self.code_gen))

    def generate_typed_dict(self, type_info: Dict[str, Any], domain_name: str) -> Dict[str, 'TypedDictGenerator']:
        """
        Generates TypedDicts based on type_info.

        Args:
            type_info: Dict containing the info for the TypedDict
            domain_name: path to resolve relative forward references in type_info against

        Returns: TypedDict corresponding to type information found in type_info
        """
        items = self._multi_fallback_get(type_info, 'returns', 'parameters', 'properties')
        td_name = self._multi_fallback_get(type_info, 'id', 'name')
        is_total = any(1 for x in items if x.get('optional'))
        td = TypedDictGenerator(td_name, is_total)
        doc_string = TypingCodeGenerator(init_imports=False)
        with td.indent_manager, doc_string.indent_manager:
            needs_closing_triple_q = False
            if 'description' in type_info or any('description' in item for item in items):
                doc_string.add_code('"""')
                needs_closing_triple_q = True
                if 'description' in type_info:
                    doc_string.add_code(type_info['description'])
                    doc_string.add_newlines(num=1)
                if any('description' in item for item in items):
                    doc_string.add_code('Attributes:')

            for item in items:
                type_ = self.convert_js_to_py_type(item, domain_name)
                if 'description' in item:
                    lines = item['description'].split('\n')
                    with doc_string.indent_manager:
                        doc_string.add_code(f'{item["name"]}: {lines[0]}')
                        if len(lines) > 1:
                            with doc_string.indent_manager:
                                doc_string.add_code(lines=lines[1:])
                td.add_code(f'{item["name"]}: {type_}')
            if needs_closing_triple_q:
                doc_string.add_code('"""')
        td.insert_code(lines=doc_string.code_lines)

        return {td_name: td}

    @staticmethod
    def _multi_fallback_get(d: Dict[Hashable, Any], *k: Hashable) -> Any:
        """
        Convenience method to retrieve item from dict with multiple keys as fallbacks for failed accesses

        Args:
            d: Dict to retrieve values from
            k: keys of Dict to retrieve values from
        Returns: Any
            first found value where key in k
        """
        for key in k:
            if key in d:
                return d[key]

        raise KeyError(f'{", ".join([str(s) for s in k])} all not found in {d}')

    def convert_js_to_py_type(self, item_info: Union[Dict[str, Any], str], domain_name) -> str:
        """
        Generates a valid python type from the JS type. In the case of type_info being a str, we simply return the
        matching python type from self.js_to_py_types. Otherwise, in the case of type_info being a Dict, we know that
        it will contain vital information about the type we are trying to convert.

        The domain_name is used to qualify relative forward reference in type_info. For example, if
        type_info['$ref'] == 'foo', domain_name would be used produce an absolute forward reference, ie domain_name.foo

        Args:
            item_info: Dict or str containing type_info
            domain_name: path to resolve relative forward references in type_info against
        Returns: str
            valid python type, either in the form of an absolute forward reference (eg Protocol.bar.foo) or
            primitive type (eg int, float, str, etc)
        """
        if isinstance(item_info, str):
            type_ = self.js_to_py_types[item_info]
        elif 'items' in item_info:
            assert item_info['type'] == 'array'
            if '$ref' in item_info['items']:
                ref = item_info['items']['$ref']
                type_ = f'List[{self.get_forward_ref(ref, domain_name)}]'
            else:
                type_ = f'List[{self.convert_js_to_py_type(item_info["items"]["type"], domain_name)}]'
        else:
            if '$ref' in item_info:
                type_ = self.get_forward_ref(item_info['$ref'], domain_name)
            else:
                if 'enum' in item_info:
                    _enum_vals = ', '.join([f'\'{x}\'' for x in item_info['enum']])
                    type_ = f'Literal[{_enum_vals}]'
                else:
                    type_ = self.js_to_py_types[item_info['type']]

        return type_

    @staticmethod
    def get_forward_ref(relative_ref: str, potential_domain_context: str) -> str:
        """
        Generates a forward absolute forward reference to Protocol class attr. If the reference is relative
        to a nested class, the full path is resolved against potential_domain_context. In the case of
        the reference being relative to the Protocol class, the path is simple resolved against the Protocol class

        Args:
            relative_ref: reference to another class, in the form of foo or foo.bar
            potential_domain_context: context to resolve class against if relative_ref is relative to it

        Returns: str
            absolute forward reference to nested class attr
        """
        if len(relative_ref.split('.')) == 2:
            non_fw_ref = f'Protocol.{relative_ref}'
        else:
            non_fw_ref = f'Protocol.{potential_domain_context}.{relative_ref}'
        return f'\'{non_fw_ref}\''


class TypingCodeGenerator:
    """
    Class to facilitate the generation of typing related related code

    Attributes:
        indent_manager: instance of IndentManager. Used to manage the current indentation level
        import_lines: List containing lines of import code
        inserted_lines: List containing lines of code before the main body
        code_lines: List containing lines of code
        lines_classification: Classification of the lines, the value should be one such that self.<value>_lines is
            defined. The default is 'code'
    """

    def __init__(self, init_imports: bool = True):
        """
        Args:
            init_imports: Whether or not to write importing code to self.import_lines, defaults to True
        """
        self.indent_manager = IndentManager()
        self.import_lines = []
        self.inserted_lines = []
        self.code_lines = []
        self.lines_classification = 'code'
        if init_imports:
            self.init_imports()

    def init_imports(self) -> None:
        """
        Writes import code relating to typing to self.import_lines
        Returns: None
        """
        self.lines_classification = 'import'
        self.add_code('import sys')
        self.add_newlines(num=1)
        self.add_code('from typing import Any, Dict, List, TYPE_CHECKING, Union')
        self.add_newlines(num=1)
        self.add_code('if sys.version_info < (3, 8):')
        with self.indent_manager:
            self.add_code('from typing_extensions import Literal, TypedDict')
        self.add_code('else:')
        with self.indent_manager:
            self.add_code('from typing import Literal, TypedDict')
        self.add_newlines(num=1)
        self.lines_classification = 'code'

    def add_newlines(self, num: int = 1, lines_classification: str = None) -> None:
        """
        Adds num newlines
        Args:
            num: number of newlines to add
            lines_classification: str of lines_classification, defaults to self.lines_classification

        Returns: None
        """
        self.add_code('\n' * num, lines_classification)

    def add_comment_from_info(self, info: Dict[str, Any]) -> None:
        """
        Adds a comment to code_lines if a description is defined in info
        Args:
            info: Dict possible containing the key info

        Returns: None
        """
        if 'description' in info:
            newline = '\n'
            self.add_code(f'# {info["description"].replace(newline, " ")}')

    def add_code(self, code: str = None, lines: List[str] = None, lines_classification: str = None) -> None:
        """
        Adds code from a string or code from lines. If code is not None, lines is assigned to the str dedented and
        split by newlines. Each line in lines is applied the current indent level before being added to the
        <lines_classification>_lines list
        Args:
            code: str containing code
            lines: List[str] containing code
            lines_classification: str of lines_classification, defaults to self.lines_classification

        Returns: None
        """
        assert code is not None or lines is not None, 'One of code or lines must be specified'
        lines_classification = lines_classification or self.lines_classification
        if code is not None:
            if all([char == '\n' for char in code]):
                # if we are adding a newline, '\n'.split('\n') == ['', ''], which will expand to 2 newlines instead of
                lines = ['' for _ in range(code.count('\n'))]
            else:
                lines = dedent(code).split('\n')
        # don't indent empty lines
        lines = [f'{self.indent_manager if li else ""}{li}' for li in lines]
        self.__getattribute__(f'{lines_classification}_lines').extend(lines)

    def __str__(self):
        return '\n'.join([*self.import_lines, *self.inserted_lines, *self.code_lines])


class TypedDictGenerator(TypingCodeGenerator):
    """
    Class to manage creation of TypedDicts specifically. Does not initialize imports.
    """

    def __init__(self, name: str, total: bool):
        super().__init__(init_imports=False)
        self.name = name
        self.total = total
        total_spec = ', total=False' if total else ''
        self.add_code(f'class {name}(TypedDict{total_spec}):')

    def filter_lines(self, *sub_pattern_replacements: Tuple[str, Any]) -> None:
        """
        Performs re.sub on all code_lines (except for the first one) according to sub_pattern_replacements. Each
        sub_sub_pattern_replacements should be a re pattern and regex replacement function/regex replacement string
        Args:
            sub_pattern_replacements: tuple containing a regex pattern and
                regex replacement function or regex replacement string

        Returns: None
        """
        # temp_code_lines = self.code_lines[:]
        # self.code_lines = [self.code_lines[0]]
        filtered_lines = []
        for index, line in enumerate(self.code_lines[1:]):
            for sub_p, sub_r in sub_pattern_replacements:
                line = re.sub(sub_p, sub_r, line)
            filtered_lines.extend(line.split('\n'))
        self.code_lines = self.code_lines[:1] + filtered_lines

    def insert_code(self, code: str = None, lines: List[str] = None, lines_classification: str = None) -> None:
        old_lines = self.code_lines[1:]
        self.code_lines = self.code_lines[:1]
        self.add_code(code, lines, lines_classification)
        self.add_code(lines=old_lines)


class IndentManager:
    """
    Simple class which can be used with a with statement to increment/decrement the current indent level/
    """

    def __init__(self):
        self.indent = ''

    def __enter__(self) -> 'IndentManager':
        self.indent += '    '
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.indent = self.indent[:-4]

    def __str__(self) -> str:
        return self.indent


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument(
        '--output',
        '-o',
        help='dir or file to output to',
        default=Path(__file__).parents[1] / 'pyppeteer' / 'models' / '_protocol.py',
    )
    generator = ProtocolTypesGenerator()
    generator.retrieve_top_level_domain()
    generator.gen_spec()
    generator.write_generated_code(path=parser.parse_args().output)



import os
import shutil
import subprocess
import time
import zipfile

import docker.types  # type: ignore

from biolib.biolib_errors import BioLibError
from biolib.compute_node.job_worker.cache_state import LfsCacheState
from biolib.compute_node.job_worker.cache_types import LargeFileSystemCache, StoragePartition
from biolib.typing_utils import TypedDict, Optional, Callable

from biolib.biolib_api_client import LargeFileSystemMapping


class StatusUpdate(TypedDict):
    progress: int
    log_message: str


class LargeFileSystemAttachResponse(TypedDict):
    aws_ebs_volume_id: str
    device_name: str


class DeviceInfo(TypedDict):
    attached_device_name: str
    aws_ebs_volume_id: str
    nvme_device_name: str


class LargeFileSystemError(BioLibError):
    pass


class LargeFileSystem:

    def __init__(
            self,
            job_id: str,
            lfs_mapping: LargeFileSystemMapping,
            send_status_update: Callable[[StatusUpdate], None],
    ):
        self._job_id: str = job_id
        self._lfs_mapping: LargeFileSystemMapping = lfs_mapping
        self._path_on_disk: Optional[str] = None
        self._send_status_update: Callable[[StatusUpdate], None] = send_status_update

    @property
    def _is_initialized(self) -> bool:
        return self._path_on_disk is not None

    @property
    def uuid(self) -> str:
        return self._lfs_mapping['uuid']

    @property
    def docker_mount(self) -> docker.types.Mount:
        if not self._is_initialized:
            raise LargeFileSystemError('LargeFileSystem not initialized')

        return docker.types.Mount(
            read_only=True,
            source=self._path_on_disk,
            target=self._lfs_mapping['to_path'],
            type='bind',
        )

    def initialize(self) -> None:
        if self._is_initialized:
            return

        lfs_size_bytes = self._lfs_mapping['size_bytes']

        lfs_is_already_downloading = False

        with LfsCacheState() as cache_state:
            lfs_cache = cache_state['large_file_systems'].get(self.uuid)

            if lfs_cache is None:

                storage_partition_to_use: Optional[StoragePartition] = None
                for storage_partition in cache_state['storage_partitions'].values():
                    free_space_bytes = storage_partition['total_size_bytes'] - storage_partition['allocated_size_bytes']
                    if lfs_size_bytes < free_space_bytes:
                        storage_partition_to_use = storage_partition
                        break

                if storage_partition_to_use is None:
                    raise LargeFileSystemError('No storage partition with space available')
                else:
                    storage_partition_to_use['allocated_size_bytes'] += lfs_size_bytes

                cache_state['large_file_systems'][self.uuid] = LargeFileSystemCache(
                    active_jobs=[self._job_id],
                    last_used_at=LfsCacheState.get_timestamp_now(),
                    size_bytes=lfs_size_bytes,
                    state='downloading',
                    storage_partition_uuid=storage_partition_to_use['uuid'],
                    uuid=self.uuid,
                )

                self._path_on_disk = f"{storage_partition_to_use['path']}/lfs/{self.uuid}/data"

            else:
                lfs_cache['active_jobs'].append(self._job_id)
                storage_partition = cache_state['storage_partitions'][lfs_cache['storage_partition_uuid']]
                self._path_on_disk = f"{storage_partition['path']}/lfs/{self.uuid}/data"

                if lfs_cache['state'] == 'ready':
                    return
                else:
                    lfs_is_already_downloading = True

        # TODO: Come up with better status reporting such that the progress values below make sense
        if lfs_is_already_downloading:
            self._send_status_update(StatusUpdate(
                progress=30,
                log_message=f'Waiting for Large File System "{self.uuid}" to be ready...',
            ))
            self._wait_for_lfs_to_be_ready()
            self._send_status_update(StatusUpdate(progress=33, log_message=f'Large File System "{self.uuid}" ready.'))
        else:
            self._send_status_update(StatusUpdate(
                progress=30,
                log_message=f'Downloading Large File System "{self.uuid}"...',
            ))
            self._download_and_unzip()
            self._send_status_update(StatusUpdate(
                progress=33,
                log_message=f'Large File System "{self.uuid}" downloaded.',
            ))
            with LfsCacheState() as cache_state:
                cache_state['large_file_systems'][self.uuid]['state'] = 'ready'

    def detach(self) -> None:
        if not self._is_initialized:
            return

        with LfsCacheState() as cache_state:
            lfs_cache = cache_state['large_file_systems'][self.uuid]
            lfs_cache['last_used_at'] = LfsCacheState.get_timestamp_now()
            lfs_cache['active_jobs'] = [job_id for job_id in lfs_cache['active_jobs'] if job_id != self._job_id]

        self._path_on_disk = None

    def _wait_for_lfs_to_be_ready(self) -> None:
        # Timeout after 15 min
        for _ in range(180):
            time.sleep(5)
            with LfsCacheState() as cache_state:
                if cache_state['large_file_systems'][self.uuid]['state'] == 'ready':
                    return

        raise LargeFileSystemError(f'Waiting for Large File System "{self.uuid}" downloading timed out')

    def _download_and_unzip(self) -> None:
        s3_lfs_bucket_name = os.environ.get('BIOLIB_S3_LFS_BUCKET_NAME')

        lfs_size_bytes = self._lfs_mapping['size_bytes']

        tmp_storage_dir: Optional[str] = None
        for path in LfsCacheState.get_tmp_storage_paths():
            disk_usage = shutil.disk_usage(path)
            if lfs_size_bytes < disk_usage.free:
                tmp_storage_dir = path

        if tmp_storage_dir is None:
            raise LargeFileSystemError('No temporary storage available for downloading Large File System')

        s3_data_zip_uri = f's3://{s3_lfs_bucket_name}/lfs/versions/{self.uuid}/data.zip'
        tmp_data_zip_path = f'{tmp_storage_dir}/lfs-{self.uuid}-data.zip'

        download_result = subprocess.run(
            ['s5cmd', 'cp', s3_data_zip_uri, tmp_data_zip_path],
            check=False,
            capture_output=True,
        )

        if download_result.returncode != 0:
            raise LargeFileSystemError(
                f'Failed to download Large File System data.zip: {download_result.stderr.decode()}'
            )

        with zipfile.ZipFile(tmp_data_zip_path, 'r') as zip_ref:
            zip_ref.extractall(self._path_on_disk)

        os.remove(tmp_data_zip_path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Helper functions."""

import asyncio
import json
import logging
import math
import re
from asyncio.futures import Future
from pathlib import Path
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, List, Optional, Union

from pyee import AsyncIOEventEmitter
from biolib.pyppeteer.pyppeteer.connection import CDPSession
from biolib.pyppeteer.pyppeteer.errors import ElementHandleError, TimeoutError

if TYPE_CHECKING:
    from pyppeteer.models import Protocol

logger = logging.getLogger(__name__)


async def future_race(*fs) -> Union[None, Exception]:
    """
    Analogous to JS's Promise.race(). Returns the results of the first completed future
    :param fs: Future to be waited upon
    :return: result of first completed Future, whether that be an exception or result

    :raises Exception: any exception raised by asyncio.wait call. Any exception raised from the first future is returned
    """
    done, _ = await asyncio.wait(fs, return_when=asyncio.FIRST_COMPLETED,)
    try:
        return done.pop().result()
    except Exception as e:
        return e


def evaluationString(fun: str, *args: Any) -> str:
    """Convert function and arguments to str."""
    _args = ', '.join([json.dumps('undefined' if arg is None else arg) for arg in args])
    expr = f'({fun})({_args})'
    return expr


def getExceptionMessage(exceptionDetails: dict) -> str:
    """Get exception message from `exceptionDetails` object."""
    exception = exceptionDetails.get('exception')
    if exception:
        return exception.get('description') or exception.get('value')
    message = exceptionDetails.get('text', '')
    stackTrace = exceptionDetails.get('stackTrace', {})
    if stackTrace:
        for callframe in stackTrace.get('callFrames'):
            location = (
                str(callframe.get('url', ''))
                + ':'
                + str(callframe.get('lineNumber', ''))
                + ':'
                + str(callframe.get('columnNumber'))
            )
            functionName = callframe.get('functionName', '<anonymous>')
            message = message + f'\n    at {functionName} ({location})'
    return message


def addEventListener(emitter: AsyncIOEventEmitter, eventName: str, handler: Callable) -> Dict[str, Any]:
    """Add handler to the emitter and return emitter/handler."""
    emitter.on(eventName, handler)
    return {'emitter': emitter, 'eventName': eventName, 'handler': handler}


def removeEventListeners(listeners: List[dict]) -> None:
    """Remove listeners from emitter."""
    for listener in listeners:
        emitter = listener['emitter']
        eventName = listener['eventName']
        handler = listener['handler']
        emitter.remove_listener(eventName, handler)
    listeners.clear()


unserializableValueMap = {
    '-0': -0,
    'NaN': None,
    None: None,
    'Infinity': math.inf,
    '-Infinity': -math.inf,
}


def valueFromRemoteObject(remoteObject: 'Protocol.Runtime.RemoteObject') -> Any:
    """Serialize value of remote object."""
    if remoteObject.get('objectId'):
        raise ElementHandleError('Cannot extract value when objectId is given')
    value = remoteObject.get('unserializableValue')
    if value:
        try:
            value = unserializableValueMap[value]
        except KeyError:
            raise ElementHandleError(f'Unsupported unserializable value: {value}')
    return remoteObject.get('value')


async def releaseObject(client: CDPSession, remoteObject: dict) -> Awaitable:
    """Release remote object."""
    objectId = remoteObject.get('objectId')
    if objectId:
        try:
            await client.send('Runtime.releaseObject', {'objectId': objectId})
        except Exception as e:
            # Exceptions might happen in case of a page been navigated or closed.
            # Swallow these since they are harmless and we don't leak anything in this case.  # noqa
            logger.error(f'An exception occurred: {e}')


def waitForEvent(
    emitter: AsyncIOEventEmitter,
    eventName: str,  # noqa: C901
    predicate: Callable[[Any], bool],
    timeout: Optional[float],
    loop: asyncio.AbstractEventLoop,
) -> Awaitable:
    """Wait for an event emitted from the emitter."""
    future = loop.create_future()

    def resolveCallback(target: Any) -> None:
        future.set_result(target)

    def rejectCallback(exception: Exception) -> None:
        future.set_exception(exception)

    async def timeoutTimer() -> None:
        if timeout:
            await asyncio.sleep(timeout / 1000)
        else:
            await asyncio.get_event_loop().create_future()
        rejectCallback(TimeoutError('Timeout exceeded while waiting for event'))

    def _listener(target: Any) -> None:
        if not predicate(target):
            return
        cleanup()
        resolveCallback(target)

    listener = addEventListener(emitter, eventName, _listener)
    if timeout:
        eventTimeout = loop.create_task(timeoutTimer())

    def cleanup() -> None:
        removeEventListeners([listener])
        if timeout:
            eventTimeout.cancel()

    return future


def get_positive_int(obj: dict, name: str) -> int:
    """Get and check the value of name in obj is positive integer."""
    value = obj[name]
    if not isinstance(value, int):
        raise TypeError(f'{name} must be integer: {type(value)}')
    elif value < 0:
        raise ValueError(f'{name} must be positive integer: {value}')
    return value


_js_identifier_re = r'(?:[\$A-Z_a-z\xAA\xB5\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0370-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386\u0388-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u05D0-\u05EA\u05F0-\u05F2\u0620-\u064A\u066E\u066F\u0671-\u06D3\u06D5\u06E5\u06E6\u06EE\u06EF\u06FA-\u06FC\u06FF\u0710\u0712-\u072F\u074D-\u07A5\u07B1\u07CA-\u07EA\u07F4\u07F5\u07FA\u0800-\u0815\u081A\u0824\u0828\u0840-\u0858\u08A0-\u08B4\u0904-\u0939\u093D\u0950\u0958-\u0961\u0971-\u0980\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BD\u09CE\u09DC\u09DD\u09DF-\u09E1\u09F0\u09F1\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A59-\u0A5C\u0A5E\u0A72-\u0A74\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABD\u0AD0\u0AE0\u0AE1\u0AF9\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3D\u0B5C\u0B5D\u0B5F-\u0B61\u0B71\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BD0\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D\u0C58-\u0C5A\u0C60\u0C61\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBD\u0CDE\u0CE0\u0CE1\u0CF1\u0CF2\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D\u0D4E\u0D5F-\u0D61\u0D7A-\u0D7F\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0E01-\u0E30\u0E32\u0E33\u0E40-\u0E46\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB0\u0EB2\u0EB3\u0EBD\u0EC0-\u0EC4\u0EC6\u0EDC-\u0EDF\u0F00\u0F40-\u0F47\u0F49-\u0F6C\u0F88-\u0F8C\u1000-\u102A\u103F\u1050-\u1055\u105A-\u105D\u1061\u1065\u1066\u106E-\u1070\u1075-\u1081\u108E\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16EE-\u16F8\u1700-\u170C\u170E-\u1711\u1720-\u1731\u1740-\u1751\u1760-\u176C\u176E-\u1770\u1780-\u17B3\u17D7\u17DC\u1820-\u1877\u1880-\u18A8\u18AA\u18B0-\u18F5\u1900-\u191E\u1950-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u1A00-\u1A16\u1A20-\u1A54\u1AA7\u1B05-\u1B33\u1B45-\u1B4B\u1B83-\u1BA0\u1BAE\u1BAF\u1BBA-\u1BE5\u1C00-\u1C23\u1C4D-\u1C4F\u1C5A-\u1C7D\u1CE9-\u1CEC\u1CEE-\u1CF1\u1CF5\u1CF6\u1D00-\u1DBF\u1E00-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u2071\u207F\u2090-\u209C\u2102\u2107\u210A-\u2113\u2115\u2118-\u211D\u2124\u2126\u2128\u212A-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2160-\u2188\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CEE\u2CF2\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D80-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u3005-\u3007\u3021-\u3029\u3031-\u3035\u3038-\u303C\u3041-\u3096\u309B-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312D\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FD5\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA61F\uA62A\uA62B\uA640-\uA66E\uA67F-\uA69D\uA6A0-\uA6EF\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AD\uA7B0-\uA7B7\uA7F7-\uA801\uA803-\uA805\uA807-\uA80A\uA80C-\uA822\uA840-\uA873\uA882-\uA8B3\uA8F2-\uA8F7\uA8FB\uA8FD\uA90A-\uA925\uA930-\uA946\uA960-\uA97C\uA984-\uA9B2\uA9CF\uA9E0-\uA9E4\uA9E6-\uA9EF\uA9FA-\uA9FE\uAA00-\uAA28\uAA40-\uAA42\uAA44-\uAA4B\uAA60-\uAA76\uAA7A\uAA7E-\uAAAF\uAAB1\uAAB5\uAAB6\uAAB9-\uAABD\uAAC0\uAAC2\uAADB-\uAADD\uAAE0-\uAAEA\uAAF2-\uAAF4\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABE2\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D\uFB1F-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE70-\uFE74\uFE76-\uFEFC\uFF21-\uFF3A\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC])(?:[\$0-9A-Z_a-z\xAA\xB5\xB7\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0300-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u0483-\u0487\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u0591-\u05BD\u05BF\u05C1\u05C2\u05C4\u05C5\u05C7\u05D0-\u05EA\u05F0-\u05F2\u0610-\u061A\u0620-\u0669\u066E-\u06D3\u06D5-\u06DC\u06DF-\u06E8\u06EA-\u06FC\u06FF\u0710-\u074A\u074D-\u07B1\u07C0-\u07F5\u07FA\u0800-\u082D\u0840-\u085B\u08A0-\u08B4\u08E3-\u0963\u0966-\u096F\u0971-\u0983\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BC-\u09C4\u09C7\u09C8\u09CB-\u09CE\u09D7\u09DC\u09DD\u09DF-\u09E3\u09E6-\u09F1\u0A01-\u0A03\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A3C\u0A3E-\u0A42\u0A47\u0A48\u0A4B-\u0A4D\u0A51\u0A59-\u0A5C\u0A5E\u0A66-\u0A75\u0A81-\u0A83\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABC-\u0AC5\u0AC7-\u0AC9\u0ACB-\u0ACD\u0AD0\u0AE0-\u0AE3\u0AE6-\u0AEF\u0AF9\u0B01-\u0B03\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3C-\u0B44\u0B47\u0B48\u0B4B-\u0B4D\u0B56\u0B57\u0B5C\u0B5D\u0B5F-\u0B63\u0B66-\u0B6F\u0B71\u0B82\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BBE-\u0BC2\u0BC6-\u0BC8\u0BCA-\u0BCD\u0BD0\u0BD7\u0BE6-\u0BEF\u0C00-\u0C03\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D-\u0C44\u0C46-\u0C48\u0C4A-\u0C4D\u0C55\u0C56\u0C58-\u0C5A\u0C60-\u0C63\u0C66-\u0C6F\u0C81-\u0C83\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBC-\u0CC4\u0CC6-\u0CC8\u0CCA-\u0CCD\u0CD5\u0CD6\u0CDE\u0CE0-\u0CE3\u0CE6-\u0CEF\u0CF1\u0CF2\u0D01-\u0D03\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D-\u0D44\u0D46-\u0D48\u0D4A-\u0D4E\u0D57\u0D5F-\u0D63\u0D66-\u0D6F\u0D7A-\u0D7F\u0D82\u0D83\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0DCA\u0DCF-\u0DD4\u0DD6\u0DD8-\u0DDF\u0DE6-\u0DEF\u0DF2\u0DF3\u0E01-\u0E3A\u0E40-\u0E4E\u0E50-\u0E59\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB9\u0EBB-\u0EBD\u0EC0-\u0EC4\u0EC6\u0EC8-\u0ECD\u0ED0-\u0ED9\u0EDC-\u0EDF\u0F00\u0F18\u0F19\u0F20-\u0F29\u0F35\u0F37\u0F39\u0F3E-\u0F47\u0F49-\u0F6C\u0F71-\u0F84\u0F86-\u0F97\u0F99-\u0FBC\u0FC6\u1000-\u1049\u1050-\u109D\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u135D-\u135F\u1369-\u1371\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16EE-\u16F8\u1700-\u170C\u170E-\u1714\u1720-\u1734\u1740-\u1753\u1760-\u176C\u176E-\u1770\u1772\u1773\u1780-\u17D3\u17D7\u17DC\u17DD\u17E0-\u17E9\u180B-\u180D\u1810-\u1819\u1820-\u1877\u1880-\u18AA\u18B0-\u18F5\u1900-\u191E\u1920-\u192B\u1930-\u193B\u1946-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u19D0-\u19DA\u1A00-\u1A1B\u1A20-\u1A5E\u1A60-\u1A7C\u1A7F-\u1A89\u1A90-\u1A99\u1AA7\u1AB0-\u1ABD\u1B00-\u1B4B\u1B50-\u1B59\u1B6B-\u1B73\u1B80-\u1BF3\u1C00-\u1C37\u1C40-\u1C49\u1C4D-\u1C7D\u1CD0-\u1CD2\u1CD4-\u1CF6\u1CF8\u1CF9\u1D00-\u1DF5\u1DFC-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u200C\u200D\u203F\u2040\u2054\u2071\u207F\u2090-\u209C\u20D0-\u20DC\u20E1\u20E5-\u20F0\u2102\u2107\u210A-\u2113\u2115\u2118-\u211D\u2124\u2126\u2128\u212A-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2160-\u2188\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D7F-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u2DE0-\u2DFF\u3005-\u3007\u3021-\u302F\u3031-\u3035\u3038-\u303C\u3041-\u3096\u3099-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312D\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FD5\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA62B\uA640-\uA66F\uA674-\uA67D\uA67F-\uA6F1\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AD\uA7B0-\uA7B7\uA7F7-\uA827\uA840-\uA873\uA880-\uA8C4\uA8D0-\uA8D9\uA8E0-\uA8F7\uA8FB\uA8FD\uA900-\uA92D\uA930-\uA953\uA960-\uA97C\uA980-\uA9C0\uA9CF-\uA9D9\uA9E0-\uA9FE\uAA00-\uAA36\uAA40-\uAA4D\uAA50-\uAA59\uAA60-\uAA76\uAA7A-\uAAC2\uAADB-\uAADD\uAAE0-\uAAEF\uAAF2-\uAAF6\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABEA\uABEC\uABED\uABF0-\uABF9\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE00-\uFE0F\uFE20-\uFE2F\uFE33\uFE34\uFE4D-\uFE4F\uFE70-\uFE74\uFE76-\uFEFC\uFF10-\uFF19\uFF21-\uFF3A\uFF3F\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC])*'


def is_js_func(func: str) -> bool:
    """
    Heuristically check if a str (when interpreted as js) is a function.
    A two step process is used.

    First, the str is stripped of whitespace with the builtin str.strip() method. Then the str is checked for whether
    or not it starts with or async, (either of which would indicate a function).

    If the str fails this check, all whitespace ( \t\r\n) is removed with str.translate. str.translate is almost an
    order of magnitude faster for large strs (tested with a 50kb str) than chained str.replace.
    Then, the larger regex is used. The regex uses a regex found here: https://stackoverflow.com/a/9337047
    and here: http://mothereff.in/js-variables to identify valid JS identifiers. (Unicode surrogates have been removed
    from the regex as they aren't supported in strs in Python)
    """
    if not isinstance(func, str):
        return False
    func = func.strip()
    if func.startswith('function ') or func.startswith('async '):
        return True
    # ~85% faster than using chained .replace()
    func = func.translate(str.maketrans(dict.fromkeys(' \t\r\n')))
    return bool(
        re.match(
            r"""
        # js functions can start with async, function, <name>, or any permutation of all three (in that order)
        # the only way a function may be named is if the function is declared with the function keyword
        (?:async)?(?P<named>function)?(?(named)(?:%(js)s)?)
        # js func args
        # matches opening ( if it exists
        (?P<argopen>\()?
        # matches one argument optionally. if function args are opening with (, we optionally match the rest argument
        (?P<args>(?(argopen)(?:\.\.\.)?)%(js)s,?)?
        # if the function opens with ( and the first arg was matched, we match more arguments,
        # otherwise, don't match any thing.  if function args are opening with (, we optionally match the rest argument
        (?(argopen) (?(args)(?:(?(argopen)(?:\.\.\.)?)%(js)s,?)*) )
        # if we matched the opening (, match the closing )
        (?(argopen)\))
        # function bodies open with =>, {, both, or none. If the function is declared with a name, it must have a { to
        # open the body. The opposite is also true. If the function has a name, it will not open with =>.
        (?(named)(?P<fnopen>{)|(?:=>))
        # asserts that there are at least 3 characters behind us (the minimum to be able create a function declaration
        # and begin declaring the function body, ie f=>).
        (?<=.{3})
        # matches the function body
        .+
        # matches the closing } if the opening one was found
        (?(fnopen)})
        """
            % {'js': _js_identifier_re},
            func,
            re.VERBOSE,
        )
    )


def safe_future_set_result(fut: Future, res: Any):
    if not fut.done():
        fut.set_result(res)


async def readProtocolStream(client: CDPSession, handle: str, path: Union[Path, str]) -> str:
    # might be better to return as bytes
    eof = False
    buffs = []
    while not eof:
        response = await client.send('IO.read', {'handle': handle})
        buffs.append(response.get('data', ''))
        eof = response.get('eof', False)
    await client.send('IO.close', {'handle': handle})

    result = ''.join(bufs)
    if path:
        Path(path).write_text(result)
    return result



import asyncio
import atexit
import json
import logging
import os
import re
import subprocess
import sys
import tempfile
import time
from contextlib import suppress
from pathlib import Path
from signal import SIG_DFL, SIGINT, SIGTERM, signal
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, Union
from urllib.error import URLError
from urllib.request import urlopen

from biolib.pyppeteer.pyppeteer import __chromium_revision__
from biolib.pyppeteer.pyppeteer.browser import Browser
from biolib.pyppeteer.pyppeteer.browser_fetcher import BrowserFetcher
from biolib.pyppeteer.pyppeteer.connection import Connection
from biolib.pyppeteer.pyppeteer.errors import BrowserError
from biolib.pyppeteer.pyppeteer.models import BrowserOptions, ChromeArgOptions, LaunchOptions, Protocol
from biolib.pyppeteer.pyppeteer.util import get_free_port
from biolib.pyppeteer.pyppeteer.websocket_transport import WebsocketTransport

if not sys.platform.startswith('win'):
    from signal import SIGHUP

if sys.version_info < (3, 8):
    pass
else:
    pass

logger = logging.getLogger(__name__)


class BrowserRunner:
    def __init__(
        self, executable_path: str, process_args: Sequence[str], temp_dir: tempfile.TemporaryDirectory = None,
    ):
        self.executable_path = executable_path
        self.process_args = process_args or []
        self.temp_dir = temp_dir

        self.proc: Optional[subprocess.Popen] = None
        self.connection: Optional[Connection] = None

        self._altered_sighandlers: Set[int] = set()
        self._closed = True

    def start(self, **kwargs: LaunchOptions) -> None:
        process_opts = {}
        if kwargs.get('pipe'):
            raise NotImplementedError('Communication via pipe not supported')
        if kwargs.get('env'):
            process_opts['env'] = kwargs['env']

        if not kwargs.get('dumpio'):
            # we read stdout to check it for the ws endpoint
            process_opts['stdout'] = subprocess.PIPE
            process_opts['stderr'] = subprocess.STDOUT
        else:
            # todo: dumpio. See: https://pptr.dev/#?product=Puppeteer&version=v2.1.1&show=api-puppeteerlaunchoptions
            # we need to tee proc stdout to both PIPE (so we can read it) and stdout (so users can see dumped IO)
            # see these SO threads:
            # https://stackoverflow.com/q/2996887/
            # https://stackoverflow.com/q/17190221/
            raise NotImplementedError(f'dumpio argument currently  not implemented')

        assert self.proc is None, 'This process has previously been started'

        logger.debug(f'Calling process: {self.executable_path} {" ".join(self.process_args)}')
        self.proc = subprocess.Popen([str(self.executable_path), *self.process_args], **process_opts)
        self._closed = False

        # ignore args from signals
        def close_proc_wrapper(*_: Any) -> None:
            asyncio.get_event_loop().run_until_complete(self._close_proc())

        if kwargs.get('autoClose'):
            atexit.register(close_proc_wrapper)
        if kwargs.get('handleSIGINT'):
            signal(SIGINT, close_proc_wrapper)
            self._altered_sighandlers.add(SIGINT)
        if kwargs.get('handleSIGTERM'):
            signal(SIGTERM, close_proc_wrapper)
            self._altered_sighandlers.add(SIGTERM)
        if kwargs.get('handleSIGHUP'):
            # SIGHUP is not defined on windows
            if not sys.platform.startswith('win'):
                signal(SIGHUP, close_proc_wrapper)
                self._altered_sighandlers.add(SIGHUP)
            else:
                logger.warning(f'SIGHUP is not available on win32')

    async def _close_proc(self) -> None:
        if not self._closed:
            if self.connection and self.connection._connected:
                await self.connection.send('Browser.close')
                await self.connection.dispose()
            if self.temp_dir:
                self._wait_for_proc_to_close()
                self.temp_dir.cleanup()

    def _wait_for_proc_to_close(self) -> None:
        if self.proc is not None and self.proc.poll() is None and not self._closed:
            try:
                self.proc.terminate()
                self.proc.wait()
            except Exception as e:
                logger.warning(f'error occurred on proc close: {e}')

    async def close(self) -> None:
        if not self._closed:
            self._restore_default_signal_handlers()
            if self.temp_dir:
                self.kill()
            elif self.connection:
                try:
                    await self.connection.send('Browser.close')
                except Exception as e:
                    logger.error(f'An exception occurred: {e}')
                    self.kill()
        return await self._close_proc()

    def kill(self) -> None:
        if self.proc and not self._closed and self.proc.returncode is not None:
            self._restore_default_signal_handlers()
            try:
                self.proc.kill()
                if sys.platform.startswith('win'):
                    subprocess.Popen(['taskkill', '/PID', str(self.proc.pid), '/F'], shell=True).communicate()

            except Exception as e:
                logger.warning(f'Exception occurred while killing process {e}')
        try:
            if self.temp_dir:
                self.temp_dir.cleanup()
        except Exception as e:
            logger.warning(f'Failed to cleanup {self.temp_dir}: {e}')

    def _restore_default_signal_handlers(self) -> None:
        """
        Restores the signals that were previously altered. This is contrary to
        restoring all handlers, which panics in multithreaded applications
        """
        for signum in self._altered_sighandlers:
            signal(signum, SIG_DFL)
        self._altered_sighandlers = set()

    async def setupConnection(
        self, usePipe: bool = None, timeout: float = None, slowMo: float = 0, preferredRevision: str = None,
    ) -> Connection:

        if usePipe:
            raise NotImplementedError('Communication via pipe not supported at this time')
            # not currently support as we have no implementation for piping the stdout
            # while also dumping to console
            # may need to transition to asyncio.subprocess
            # transport = PipeTransport(write_stream, read_stream)
            # self.connection = Connection('', transport, delay=slowMo)
        if self.proc is None:
            raise RuntimeError('class process not initialized (self.proc = None)')
        browser_ws_endpoint = waitForWSEndpoint(self.proc, timeout, preferredRevision)
        transport = await WebsocketTransport.create(uri=browser_ws_endpoint)
        self.connection = Connection(url=browser_ws_endpoint, transport=transport, delay=slowMo)
        return self.connection


class BaseBrowserLauncher:
    """
    Implements common BrowserLauncher operations
    """

    def __init__(self, projectRoot: Union[Path, str] = None, preferredRevision: str = None):
        self.projectRoot = Path(projectRoot) if projectRoot else None
        self.preferredRevision = preferredRevision

    async def connect(
        self,
        browserWSEndpoint: str = None,
        browserURL: str = None,
        transport: WebsocketTransport = None,
        ignoreHTTPSErrors: bool = False,
        slowMo: float = 0,
        defaultViewport: Protocol.Page.Viewport = None,
    ) -> Browser:
        if defaultViewport is None:
            defaultViewport = {'width': 800, 'height': 600}

        assert (
            len([x for x in (browserWSEndpoint, browserURL, transport) if x]) == 1
        ), 'exactly one of browserWSEndpoint, browserURL, and transport must be specified'

        if transport:
            connection = Connection('', transport, slowMo)
        else:
            if browserURL:
                browserWSEndpoint = getWSEndpoint(browserURL)
            transport = await WebsocketTransport.create(uri=browserWSEndpoint)
            connection = Connection(browserWSEndpoint, transport=transport, delay=slowMo)

        async def close_callback() -> None:
            await connection.send('Browser.close')

        context_ids = await connection.send('Target.getBrowserContexts')
        return await Browser.create(
            connection=connection,
            contextIds=context_ids,
            ignoreHTTPSErrors=ignoreHTTPSErrors,
            defaultViewport=defaultViewport,
            process=None,
            closeCallback=close_callback,
        )


class ChromeLauncher(BaseBrowserLauncher):
    DEFAULT_ARGS = [
        '--disable-background-networking',
        '--enable-features=NetworkService,NetworkServiceInProcess',
        '--disable-background-timer-throttling',
        '--disable-backgrounding-occluded-windows',
        '--disable-breakpad',
        '--disable-client-side-phishing-detection',
        '--disable-component-extensions-with-background-pages',
        '--disable-default-apps',
        '--disable-dev-shm-usage',
        '--disable-extensions',
        '--disable-features=TranslateUI',
        '--disable-hang-monitor',
        '--disable-ipc-flooding-protection',
        '--disable-popup-blocking',
        '--disable-prompt-on-repost',
        '--disable-renderer-backgrounding',
        '--disable-sync',
        '--force-color-profile=srgb',
        '--metrics-recording-only',
        '--no-first-run',
        '--enable-automation',
        '--password-store=basic',
        '--use-mock-keychain',
    ]
    product = 'chrome'

    def __init__(self, projectRoot: str = None, preferredRevision: str = None):
        if not preferredRevision:
            preferredRevision = __chromium_revision__
        super().__init__(projectRoot, preferredRevision)

    @property
    def executable_path(self) -> Optional[str]:
        return resolveExecutablePath(self.projectRoot, self.preferredRevision)[0]

    async def launch(self, **kwargs: Union[LaunchOptions, ChromeArgOptions, BrowserOptions]) -> Browser:
        ignoreDefaultArgs = kwargs.get('ignoreDefaultArgs', False)
        args: Sequence[str] = kwargs.get('args', [])
        dumpio = kwargs.get('dumpio', False)
        executablePath = kwargs.get('executablePath', None)
        env = kwargs.get('env', os.environ)
        handleSIGINT = kwargs.get('handleSIGINT', True)
        handleSIGTERM = kwargs.get('handleSIGTERM', True)
        handleSIGHUP = kwargs.get('handleSIGHUP', not sys.platform.startswith('win'))
        ignoreHTTPSErrors = kwargs.get('ignoreHTTPSErrors', False)
        defaultViewport = kwargs.get('defaultViewport', {'width': 800, 'height': 600})
        slowMo = kwargs.get('slowMo', 0)
        timeout = kwargs.get('timeout', 30_000)
        profile_path = None

        chrome_args = []
        if not ignoreDefaultArgs:
            chrome_args.extend(self.default_args(**kwargs))
        elif isinstance(ignoreDefaultArgs, list):
            chrome_args.extend([x for x in self.default_args(**kwargs) if x not in ignoreDefaultArgs])
        else:
            chrome_args.extend(args)

        if not any(x.startswith('--remote-debugging-') for x in chrome_args):
            chrome_args.append(f'--remote-debugging-port={get_free_port()}')
        if not any(x.startswith(f'--user-data-dir') for x in chrome_args):
            profile_path = tempfile.TemporaryDirectory(prefix='pyppeteer_chrome_profile_')
            chrome_args.append(f'--user-data-dir={profile_path.name}')

        if not executablePath:
            chrome_executable, missing_text = resolveExecutablePath(self.projectRoot, self.preferredRevision)
            if missing_text:
                raise RuntimeError(missing_text)
        else:
            chrome_executable = executablePath

        usePipe = '--remote-debugging-pipe' in chrome_args

        runner = BrowserRunner(chrome_executable, chrome_args, profile_path)
        runner.start(
            handleSIGINT=handleSIGINT, handleSIGHUP=handleSIGHUP, handleSIGTERM=handleSIGTERM, env=env, dumpio=dumpio, autoClose=kwargs.get('autoClose')
        )

        try:
            connection = await runner.setupConnection(
                usePipe=usePipe, timeout=timeout, slowMo=slowMo, preferredRevision=self.preferredRevision,
            )
            browser = await Browser.create(
                connection=connection,
                contextIds=[],
                ignoreHTTPSErrors=ignoreHTTPSErrors,
                defaultViewport=defaultViewport,
                process=runner.proc,
                closeCallback=runner.close,
            )
            await browser.waitForTarget(lambda x: x.type == 'page')
            return browser
        except Exception as original_exception:
            logger.error(f'error occurred while trying to launch browser: {original_exception}')
            with suppress(Exception):
                runner.kill()
            raise original_exception

    def default_args(
        self,
        args: Sequence[str] = None,
        devtools: bool = False,
        headless: bool = None,
        userDataDir: str = None,
        **_: Any,
    ) -> List[str]:
        if headless is None:
            headless = not devtools
        if args is None:
            args = []
        chrome_args = self.DEFAULT_ARGS[:]
        if isinstance(args, Sequence):
            chrome_args.extend(args)
        if userDataDir:
            chrome_args.append(f'--user-data-dir={userDataDir}')
        if devtools:
            chrome_args.append('--auto-open-devtools-for-tabs')
        if headless:
            chrome_args.extend(('--headless', '--hide-scrollbars', '--mute-audio'))
        if all(x.startswith('-') for x in args):
            chrome_args.append('about:blank')
        return chrome_args


class FirefoxLauncher(BaseBrowserLauncher):
    DEFAULT_ARGS = [
        '--remote-debugging-port=0',
        '--no-remote',
        '--foreground',
    ]
    _server = 'dummy.test'
    DEFAULT_PROFILE_PREFS = {
        # Make sure Shield doesn't hit the network.
        'app.normandy.api_url': '',
        # Disable Firefox old build background check
        'app.update.checkInstallTime': False,
        # Disable automatically upgrading Firefox
        'app.update.disabledForTesting': True,
        # Increase the APZ content response timeout to 2 minute
        'apz.content_response_timeout': 60000,
        # Prevent various error message on the console
        # jest-puppeteer asserts that no error message is emitted by the console
        'browser.contentblocking.features.standard': '-tp,tpPrivate,cookieBehavior0,-cm,-fp',
        # Enable the dump function: which sends messages to the system
        # console
        # https://bugzilla.mozilla.org/show_bug.cgi?id=1543115
        'browser.dom.window.dump.enabled': True,
        # Disable topstories
        'browser.newtabpage.activity-stream.feeds.section.topstories': False,
        # Always display a blank page
        'browser.newtabpage.enabled': False,
        # Background thumbnails in particular cause grief: and disabling
        # thumbnails in general cannot hurt
        'browser.pagethumbnails.capturing_disabled': True,
        # Disable safebrowsing components.
        'browser.safebrowsing.blockedURIs.enabled': False,
        'browser.safebrowsing.downloads.enabled': False,
        'browser.safebrowsing.malware.enabled': False,
        'browser.safebrowsing.passwords.enabled': False,
        'browser.safebrowsing.phishing.enabled': False,
        # Disable updates to search engines.
        'browser.search.update': False,
        # Do not restore the last open set of tabs if the browser has crashed
        'browser.sessionstore.resume_from_crash': False,
        # Skip check for default browser on startup
        'browser.shell.checkDefaultBrowser': False,
        # Disable newtabpage
        'browser.startup.homepage': 'about:blank',
        # Do not redirect user when a milstone upgrade of Firefox is detected
        'browser.startup.homepage_override.mstone': 'ignore',
        # Start with a blank page about:blank
        'browser.startup.page': 0,
        # Do not allow background tabs to be zombified on Android: otherwise for
        # tests that open additional tabs: the test harness tab itself might get
        # unloaded
        'browser.tabs.disableBackgroundZombification': False,
        # Do not warn when closing all other open tabs
        'browser.tabs.warnOnCloseOtherTabs': False,
        # Do not warn when multiple tabs will be opened
        'browser.tabs.warnOnOpen': False,
        # Disable the UI tour.
        'browser.uitour.enabled': False,
        # Turn off search suggestions in the location bar so as not to trigger
        # network connections.
        'browser.urlbar.suggest.searches': False,
        # Disable first run splash page on Windows 10
        'browser.usedOnWindows10.introURL': '',
        # Do not warn on quitting Firefox
        'browser.warnOnQuit': False,
        # Do not show datareporting policy notifications which can
        # interfere with tests
        'datareporting.healthreport.about.reportUrl': f'http://{_server}/dummy/abouthealthreport/',
        'datareporting.healthreport.documentServerURI': f'http://{_server}/dummy/healthreport/',
        'datareporting.healthreport.logging.consoleEnabled': False,
        'datareporting.healthreport.service.enabled': False,
        'datareporting.healthreport.service.firstRun': False,
        'datareporting.healthreport.uploadEnabled': False,
        'datareporting.policy.dataSubmissionEnabled': False,
        'datareporting.policy.dataSubmissionPolicyAccepted': False,
        'datareporting.policy.dataSubmissionPolicyBypassNotification': True,
        # DevTools JSONViewer sometimes fails to load dependencies with its require.js.
        # This doesn't affect Puppeteer but spams console (Bug 1424372)
        'devtools.jsonview.enabled': False,
        # Disable popup-blocker
        'dom.disable_open_during_load': False,
        # Enable the support for File object creation in the content process
        # Required for |Page.setFileInputFiles| protocol method.
        'dom.file.createInChild': True,
        # Disable the ProcessHangMonitor
        'dom.ipc.reportProcessHangs': False,
        # Disable slow script dialogues
        'dom.max_chrome_script_run_time': 0,
        'dom.max_script_run_time': 0,
        # Only load extensions from the application and user profile
        # AddonManager.SCOPE_PROFILE + AddonManager.SCOPE_APPLICATION
        'extensions.autoDisableScopes': 0,
        'extensions.enabledScopes': 5,
        # Disable metadata caching for installed add-ons by default
        'extensions.getAddons.cache.enabled': False,
        # Disable installing any distribution extensions or add-ons.
        'extensions.installDistroAddons': False,
        # Disabled screenshots extension
        'extensions.screenshots.disabled': True,
        # Turn off extension updates so they do not bother tests
        'extensions.update.enabled': False,
        # Turn off extension updates so they do not bother tests
        'extensions.update.notifyUser': False,
        # Make sure opening about:addons will not hit the network
        'extensions.webservice.discoverURL': f'http://{_server}/dummy/discoveryURL',
        # Allow the application to have focus even it runs in the background
        'focusmanager.testmode': True,
        # Disable useragent updates
        'general.useragent.updates.enabled': False,
        # Always use network provider for geolocation tests so we bypass the
        # macOS dialog raised by the corelocation provider
        'geo.provider.testing': True,
        # Do not scan Wifi
        'geo.wifi.scan': False,
        # No hang monitor
        'hangmonitor.timeout': 0,
        # Show chrome errors and warnings in the error console
        'javascript.options.showInConsole': True,
        # Disable download and usage of OpenH264: and Widevine plugins
        'media.gmp-manager.updateEnabled': False,
        # Prevent various error message on the console
        # jest-puppeteer asserts that no error message is emitted by the console
        'network.cookie.cookieBehavior': 0,
        # Do not prompt for temporary redirects
        'network.http.prompt-temp-redirect': False,
        # Disable speculative connections so they are not reported as leaking
        # when they are hanging around
        'network.http.speculative-parallel-limit': 0,
        # Do not automatically switch between offline and online
        'network.manage-offline-status': False,
        # Make sure SNTP requests do not hit the network
        'network.sntp.pools': _server,
        # Disable Flash.
        'plugin.state.flash': 0,
        'privacy.trackingprotection.enabled': False,
        # Enable Remote Agent
        # https://bugzilla.mozilla.org/show_bug.cgi?id=1544393
        'remote.enabled': True,
        # Don't do network connections for mitm priming
        'security.certerrors.mitm.priming.enabled': False,
        # Local documents have access to all other local documents,
        # including directory listings
        'security.fileuri.strict_origin_policy': False,
        # Do not wait for the notification button security delay
        'security.notification_enable_delay': 0,
        # Ensure blocklist updates do not hit the network
        'services.settings.server': f'http://{_server}/dummy/blocklist/',
        # Do not automatically fill sign-in forms with known usernames and
        # passwords
        'signon.autofillForms': False,
        # Disable password capture, so that tests that include forms are not
        # influenced by the presence of the persistent doorhanger notification
        'signon.rememberSignons': False,
        # Disable first-run welcome page
        'startup.homepage_welcome_url': 'about:blank',
        # Disable first-run welcome page
        'startup.homepage_welcome_url.additional': '',
        # Disable browser animations (tabs, fullscreen, sliding alerts)
        'toolkit.cosmeticAnimations.enabled': False,
        # We want to collect telemetry, but we don't want to send in the results
        'toolkit.telemetry.server': f'https://{_server}/dummy/telemetry/',
        # Prevent starting into safe mode after application crashes
        'toolkit.startup.max_resumed_crashes': -1,
    }
    product = 'firefox'

    def __init__(self, projectRoot: str = None, preferredRevision: str = None):
        super().__init__(projectRoot, preferredRevision)

    @property
    def executablePath(self) -> Optional[Path]:
        raise NotImplementedError('executablePath method not implemented')

    async def launch(self, **kwargs: Union[LaunchOptions, ChromeArgOptions, BrowserOptions]) -> Browser:
        ignoreDefaultArgs = kwargs.get('ignoreDefaultArgs', False)
        args: Sequence[str] = kwargs.get('args', [])
        dumpio = kwargs.get('dumpio', False)
        executablePath = kwargs.get('executablePath', None)
        pipe = kwargs.get('pipe', False)
        env = kwargs.get('env', os.environ)
        handleSIGINT = kwargs.get('handleSIGINT', True)
        handleSIGTERM = kwargs.get('handleSIGTERM', True)
        handleSIGHUP = kwargs.get('handleSIGHUP', not sys.platform.startswith('win'))
        ignoreHTTPSErrors = kwargs.get('ignoreHTTPSErrors', False)
        defaultViewport = kwargs.get('defaultViewport', {'width': 800, 'height': 600})
        slowMo = kwargs.get('slowMo', 0)
        timeout = kwargs.get('timeout', 30_000)
        profile_path = None

        firefox_args = []
        if not ignoreDefaultArgs:
            firefox_args.extend(self.default_args(**kwargs))
        elif isinstance(ignoreDefaultArgs, Sequence):
            firefox_args.extend([x for x in self.default_args()])
        else:
            firefox_args.extend(args)

        if '-profile' not in firefox_args and '--profile' not in firefox_args:
            profile_path = tempfile.TemporaryDirectory('pyppyeteer2_firefox_profile_')
            firefox_args.extend(('--profile', profile_path.name))

        if not executablePath:
            missing_text, executablePath = resolveExecutablePath(self.projectRoot, self.preferredRevision)
            if missing_text:
                raise RuntimeError(missing_text)

        runner = BrowserRunner(executable_path=executablePath, process_args=firefox_args, temp_dir=profile_path)
        runner.start(
            handleSIGHUP=handleSIGHUP,
            handleSIGTERM=handleSIGTERM,
            handleSIGINT=handleSIGINT,
            dumpio=dumpio,
            env=env,
            pipe=pipe,
        )

        try:
            connection = await runner.setupConnection(
                usePipe=pipe, timeout=timeout, slowMo=slowMo, preferredRevision=self.preferredRevision,
            )
            browser = await Browser.create(
                connection=connection,
                contextIds=[],
                ignoreHTTPSErrors=ignoreHTTPSErrors,
                defaultViewport=defaultViewport,
                process=runner.proc,
                closeCallback=runner.close,
            )
            await browser.waitForTarget(lambda t: t.type == 'page')
            return browser
        except Exception as original_exception:
            logger.error(f'error occurred while trying to launch browser: {original_exception}')
            with suppress(Exception):
                runner.kill()
            raise original_exception

    def default_args(
        self,
        args: Sequence[str] = None,
        devtools: bool = False,
        headless: bool = None,
        userDataDir: str = None,
        **_: Any,
    ) -> List[str]:
        if headless is None:
            headless = not devtools
        proc_args = self.DEFAULT_ARGS[:]
        if isinstance(args, Sequence):
            proc_args.extend(args)
        if userDataDir:
            proc_args.extend(['--profile', userDataDir])
        if headless:
            proc_args.append('--headless')
        if devtools:
            proc_args.append('--devtools')
        if all(x.startswith('-') for x in proc_args):
            proc_args.append('about:blank')

        return proc_args

    def _create_profile(self, extra_prefs: Dict[str, Any]) -> tempfile.TemporaryDirectory:
        profile_path = tempfile.TemporaryDirectory(prefix='pyppeteer_firefox_profile')
        prefs = {**self.DEFAULT_PROFILE_PREFS, **extra_prefs}
        serialized_prefs = [f'user_pref({json.dumps(key)}, {json.dumps(val)}' for key, val in prefs.items()]
        with open(Path(profile_path.name).joinpath('user.js'), 'w') as user:
            user.write('\n'.join(serialized_prefs))
        return profile_path


def waitForWSEndpoint(proc: subprocess.Popen, timeout: Optional[float], preferredRevision: Optional[str]) -> str:
    assert proc.stdout is not None, 'process STDOUT wasn\'t piped'
    start = time.perf_counter()
    buffer = ''
    for line in iter(proc.stdout.readline, b''):
        line = line.decode()
        buffer += '\n' + line
        if timeout and (start - time.perf_counter()) > timeout:
            raise TimeoutError(
                f'Timed out after {timeout * 1000:.0f}ms while trying to connect to the browser! '
                f'Only Chrome at revision {preferredRevision} is guaranteed to work.'
            )
        potential_match = re.match(r'DevTools listening on (ws://[\w.:/-]*)+', line)
        if potential_match:
            return potential_match.group(1)
    raise RuntimeError(
        buffer + '\nProcess ended before WebSockets endpoint could be found.'
        f'Only Chrome at revision {preferredRevision} is guaranteed to work.'
    )


def getWSEndpoint(url: str) -> str:
    url += '/json/version'
    timeout = time.perf_counter() + 30
    while True:
        if time.perf_counter() > timeout:
            raise BrowserError('Browser closed unexpectedly:\n')
        try:
            with urlopen(url) as f:
                data = json.loads(f.read().decode())
            break
        except URLError:
            time.sleep(0.1)
    try:
        return data['webSocketDebuggerUrl']
    except KeyError:
        raise RuntimeError(f'webSocketDebuggerUrl not found')


def resolveExecutablePath(projectRoot: Path, preferred_revision: str) -> Tuple[Optional[Path], Optional[str]]:
    missing_text = None
    exec_path_env_var = 'PYPPETEER_EXECUTABLE_PATH'
    revision_env_var = 'PYPPETEER_CHROMIUM_REVISION'
    executable = os.environ.get(exec_path_env_var)
    if executable:
        if not Path(executable).is_file():
            missing_text = f'Tried to use env variable ({exec_path_env_var}) to launch browser, but no executable was found at {executable}'
            return None, missing_text
    browser_fetcher = BrowserFetcher(projectRoot)
    revision = os.environ.get(revision_env_var)
    if revision:
        revision_info = browser_fetcher.revision_info(revision)
        if not revision_info['local']:
            missing_text = f'Tried to use env variables ({revision_env_var}) to launch browser, but did not find executable at {revision_info["executablePath"]}'
            return None, missing_text
    revision_info = browser_fetcher.revision_info(preferred_revision)
    if not revision_info['local']:
        missing_text = 'Browser is not downloaded. Try running pyppeteer-install'
    return revision_info['executablePath'], missing_text


def launcher(
    projectRoot: str = None, preferredRevision: str = None, product: str = None
) -> Union[FirefoxLauncher, ChromeLauncher]:
    """Returns the appropriate browser launcher class instance"""
    product = product or os.environ.get('PYPPETEER_PRODUCT') or 'chrome'
    if product == 'firefox':
        return FirefoxLauncher(projectRoot, preferredRevision)
    elif product in ('chrome', 'chromium'):
        return ChromeLauncher(projectRoot, preferredRevision)
    raise ValueError(f'Support for {product} has not been implemented')



import base64
import random
import sys
import time
import threading
import socket
from queue import Queue

from biolib.compute_node.job_worker import JobWorkerProcess
from biolib.compute_node.socker_listener_thread import SocketListenerThread
from biolib.compute_node.socket_sender_thread import SocketSenderThread
from biolib.compute_node.webserver import webserver_utils
from biolib.biolib_binary_format import AttestationDocument, SystemStatusUpdate, SystemException
from biolib.compute_node.utils import get_package_type, WorkerThreadException, SystemExceptionCodes
from biolib.biolib_logging import logger

SOCKET_HOST = '127.0.0.1'


class WorkerThread(threading.Thread):
    def __init__(self, compute_state):
        try:
            super().__init__()
            self.compute_state = compute_state
            self._socket_port = random.choice(range(6000, 65000))
            self._socket = None
            self._connection = None
            self._job_worker_process = None
            self._connection_thread = None
            self._listener_thread = None
            self._sender_thread = None
            self._start_and_connect_to_compute_process()

            logger.debug(f"WorkerThread connected to port {self._socket_port}")

        except Exception as exception:
            raise WorkerThreadException(exception, SystemExceptionCodes.FAILED_TO_INITIALIZE_WORKER_THREAD.value,
                                        worker_thread=self) from exception

    def run(self):
        try:
            while True:
                package = self.compute_state['received_messages_queue'].get()
                package_type = get_package_type(package)

                if package_type == 'AttestationDocument':
                    self.compute_state['attestation_document'] = AttestationDocument(package).deserialize()

                elif package_type == 'StdoutAndStderr':
                    self.compute_state['status']['stdout_and_stderr_packages_b64'].append(
                        base64.b64encode(package).decode()
                    )

                elif package_type == 'SystemStatusUpdate':
                    progress, log_message = SystemStatusUpdate(package).deserialize()
                    self.compute_state['status']['status_updates'].append({'progress': progress,
                                                                           'log_message': log_message})
                    self.compute_state['progress'] = progress

                elif package_type == 'SystemException':
                    error_code = SystemException(package).deserialize()
                    self.compute_state['status']['error_code'] = error_code
                    logger.debug("Hit error. Terminating Worker Thread and Compute Process")
                    self.terminate()

                elif package_type == 'ModuleOutput':
                    self.compute_state['result'] = package
                    self.terminate()

                elif package_type == 'AesEncryptedPackage':
                    if self.compute_state['progress'] == 95:  # Check if encrypted package is ModuleOutput
                        self.compute_state['result'] = package
                        self.terminate()
                    else:  # Else it is StdoutAndStderr
                        self.compute_state['status']['stdout_and_stderr_packages_b64'].append(
                            base64.b64encode(package).decode()
                        )

                else:
                    raise Exception(f'Package type from child was not recognized: {package}')

                self.compute_state['received_messages_queue'].task_done()

        except Exception as exception:
            raise WorkerThreadException(exception, SystemExceptionCodes.FAILED_TO_HANDLE_PACKAGE_IN_WORKER_THREAD.value,
                                        worker_thread=self) from exception

    def _start_and_connect_to_compute_process(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        logger.debug(f"Trying to bind to socket on {SOCKET_HOST}:{self._socket_port}")
        self._socket.bind((SOCKET_HOST, self._socket_port))

        logger.debug(f"Starting to listen to socket on port {self._socket_port}")
        self._socket.listen()
        logger.debug(f"Listening to port {self._socket_port}")

        received_messages_queue = Queue()
        messages_to_send_queue = Queue()

        # Starting a thread for accepting connections before starting the process that should to connect to the socket
        logger.debug("Starting connection thread")
        self._connection_thread = threading.Thread(target=self._accept_new_socket_connection, args=[
            received_messages_queue,
            messages_to_send_queue
        ])
        self._connection_thread.start()
        logger.debug("Started connection thread")
        logger.debug("Starting compute process")

        self._job_worker_process = JobWorkerProcess(socket_port=self._socket_port, log_level=logger.level)
        self._job_worker_process.start()

        self.compute_state['received_messages_queue'] = received_messages_queue
        self.compute_state['messages_to_send_queue'] = messages_to_send_queue
        self.compute_state['worker_thread'] = self

    def _accept_new_socket_connection(self, received_messages_queue, messages_to_send_queue):
        self._connection, _ = self._socket.accept()
        self._listener_thread = SocketListenerThread(self._connection, received_messages_queue)
        self._listener_thread.start()

        self._sender_thread = SocketSenderThread(self._connection, messages_to_send_queue)
        self._sender_thread.start()

    def terminate(self):
        if self._job_worker_process:
            logger.debug(f'Terminating JobWorkerProcess with PID {self._job_worker_process.pid}')
            self._job_worker_process.terminate()

        if self._socket:
            self._socket.close()

        if self._connection:
            self._connection.close()

        if self.compute_state['result']:
            seconds_to_sleep = 60
            job_id = self.compute_state['job_id']
            logger.debug(f'Worker thread sleeping for {seconds_to_sleep} seconds before cleaning up job {job_id}')
            # sleep to see if the user has begun downloading the result
            time.sleep(seconds_to_sleep)
            if self.compute_state['result']:
                logger.debug(f"Job {job_id}'s result was not fetched within {seconds_to_sleep}s, cleaning up...")

        # clean up job
        webserver_utils.finalize_and_clean_up_compute_job(job_id)

        logger.debug("Terminating Worker Thread")
        sys.exit()
